(self.webpackChunkmitosheet=self.webpackChunkmitosheet||[]).push([[568],{1568:(e,l,a)=>{"use strict";a.r(l),a.d(l,{MODULE_NAME:()=>t.o,MODULE_VERSION:()=>t.Y,ExampleModel:()=>s.ExampleModel,ExampleView:()=>s.ExampleView,default:()=>E});var t=a(8657),s=a(5361);const E=a(4480).default}}]);